#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Function action
============================================================================

The :class:`Function` action wraps a callable, optionally with some
default keyword argument values.  On execution, the execution data
(commonly containing the recognition extras) are combined with the
default argument values (if present) to form the arguments with which
the callable will be called.

Simple usage::

    >>> def func(count):
    ...     print("count: %d" % count)
    ...
    >>> action = Function(func)
    >>> action.execute({"count": 2})
    count: 2
    True
    >>> # Additional keyword arguments are ignored:
    >>> action.execute({"count": 2, "flavor": "vanilla"})
    count: 2
    True

Usage with default arguments::

    >>> def func(count, flavor):
    ...     print("count: %d" % count)
    ...     print("flavor: %s" % flavor)
    ...
    >>> # The Function object can be given default argument values:
    >>> action = Function(func, flavor="spearmint")
    >>> action.execute({"count": 2})
    count: 2
    flavor: spearmint
    True
    >>> # Arguments given at the execution-time to override default values:
    >>> action.execute({"count": 2, "flavor": "vanilla"})
    count: 2
    flavor: vanilla
    True

Usage with the ``remap_data`` argument::

    >>> def func(x, y, z):
    ...     print("x: %d" % x)
    ...     print("y: %d" % y)
    ...     print("z: %d" % z)
    ...
    >>> # The Function object can optionally be given a second dictionary
    >>> # argument to use extras with different names. It should be
    >>> # compatible with the 'defaults' parameter:
    >>> action = Function(func, dict(n="x", m="y"), z=4)
    >>> action.execute({"n": 2, "m": 3})
    x: 2
    y: 3
    z: 4
    True


Class reference
----------------------------------------------------------------------------

"""

import inspect

import six

from dragonfly.actions.action_base      import ActionBase, ActionError


#---------------------------------------------------------------------------

class Function(ActionBase):
    """ Call a function with extra keyword arguments. """

    def __init__(self, function, remap_data=None, **defaults):
        """
            Constructor arguments:
             - *function* (callable) --
               the function to call when this action is executed
             - *remap_data* (dict, default: None) --
               optional dict of data keys to function keyword arguments
             - defaults --
               default keyword-values for the arguments with which
               the function will be called

        """
        ActionBase.__init__(self)
        self._function = function
        self._defaults = defaults
        self._remap_data = remap_data or {}
        self._str = function.__name__

        # Get argument names and defaults. Use getfullargspec() in Python 3
        # to avoid deprecation warnings.
        if six.PY2:
            # pylint: disable=deprecated-method
            argspec = inspect.getargspec(self._function)
        else:
            argspec = inspect.getfullargspec(self._function)

        args, varkw = argspec[0], argspec[2]
        self._filter_keywords = not varkw
        self._valid_keywords = set(args)

    def _execute(self, data=None):
        arguments = dict(self._defaults)
        if isinstance(data, dict):
            arguments.update(data)

        # Remap specified names.
        if arguments and self._remap_data:
            for old_name, new_name in self._remap_data.items():
                if old_name in data:
                    arguments[new_name] = arguments.pop(old_name)

        if self._filter_keywords:
            invalid_keywords = set(arguments.keys()) - self._valid_keywords
            for key in invalid_keywords:
                del arguments[key]

        try:
            self._function(**arguments)
        except Exception as e:
            self._log.exception("Exception from function %s:",
                                self._function.__name__)
            raise ActionError("%s: %s" % (self, e))

    def __str__(self):
        if (self._str == '<lambda>'):
            try:
                return '{!r}()'.format(inspect.getsource(self._function)
                                       .strip())
            except (OSError, IOError):
                pass
        return '{!r}()'.format(self._str)



# This file was part of Aenea
#
# Aenea is free software: you can redistribute it and/or modify it under
# the terms of version 3 of the GNU Lesser General Public License as
# published by the Free Software Foundation.
#
# Aenea is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with Aenea.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright (2014) Alex Roper
# Alex Roper <alex@aroper.net>

# This file is based on Aenea's X11 key_press implementations.

# pylint: disable=abstract-method

from locale import getpreferredencoding

from six import binary_type
from ._base import BaseKeyboard, BaseTypeable, BaseKeySymbols


KEY_TRANSLATION = {
    # Dragonfly references character keys as characters internally.
    '&': 'ampersand',
    '*': 'asterisk',
    '@': 'at',
    '\\': 'backslash',
    '`': 'grave',
    '|': 'bar',
    '^': 'asciicircum',
    ':': 'colon',
    ',': 'comma',
    '$': 'dollar',
    '.': 'period',
    '"': 'quotedbl',
    '=': 'equal',
    '!': 'exclam',
    '#': 'numbersign',
    '-': 'minus',
    '<': 'less',
    '{': 'braceleft',
    '[': 'bracketleft',
    '(': 'parenleft',
    '%': 'percent',
    '+': 'plus',
    '?': 'question',
    '>': 'greater',
    '}': 'braceright',
    ']': 'bracketright',
    ')': 'parenright',
    ';': 'semicolon',
    '/': 'slash',
    "'": 'apostrophe',
    '~': 'asciitilde',
    '_': 'underscore',
    ' ': 'space',
}


def _update_key_translation(translation):
    caps_keys = [
        'left',
        'right',
        'up',
        'down',
        'home',
        'end',
        'tab',
        'insert',
        'escape'
        ]
    caps_keys.extend('f%i' % i for i in range(1, 13))
    for key in caps_keys:
        translation[key] = key[0].upper() + key[1:]
    for index in range(10):
        translation['np%i' % index] = 'KP_%i' % index
    letters = tuple(range(ord('a'), ord('z')))
    digits = tuple(range(ord('0'), ord('9')))
    for char in letters + digits:
        translation[chr(char)] = chr(char)
        translation[chr(char).upper()] = chr(char).upper()


_update_key_translation(KEY_TRANSLATION)


class X11Typeable(BaseTypeable):
    pass


class BaseX11Keyboard(BaseKeyboard):
    """ Base Keyboard class for X11. """

    @classmethod
    def get_typeable(cls, char, is_text=False):
        """ Get a Typeable object. """
        # Get a key translation if one exists.  Otherwise use the character
        #  as the key name.
        key = KEY_TRANSLATION.get(char, char)
        if isinstance(key, binary_type):
            key = key.decode(getpreferredencoding())

        # Convert single character keys to their Unicode code points.
        #  This allows typing any Unicode character with the Text and Key
        #  actions.  It works with both X11 keyboard implementations.
        if len(key) == 1:
            # Get the Unicode code point for the character.
            encoding = getpreferredencoding()
            code_point = key.encode("unicode_escape")[2:].decode(encoding)

            # Handle ASCII keys by getting the hex code without '0x'.
            if not code_point:
                code_point = hex(ord(key))[2:]

            # Create a key name that xdotool will accept (e.g. U20AC).
            key = 'U' + code_point.upper()
        return X11Typeable(code=key, name=char, is_text=is_text)


class XdoKeySymbols(BaseKeySymbols):
    """ Key symbols for libxdo. """

    # Whitespace and editing keys
    RETURN = 'Return'
    TAB = 'Tab'
    SPACE = 'space'
    BACK = 'BackSpace'
    DELETE = 'Delete'

    # Main modifier keys
    SHIFT = 'Shift_L'
    CONTROL = 'Control_L'
    ALT = 'Alt_L'

    # Right modifier keys
    RSHIFT = 'Shift_R'
    RCONTROL = 'Control_R'
    RALT = 'Alt_R'

    # Special keys
    ESCAPE = 'Escape'
    INSERT = 'Insert'
    PAUSE = 'Pause'
    LSUPER = 'Super_L'
    RSUPER = 'Super_R'
    APPS = 'Menu'
    SNAPSHOT = 'Print'

    # Lock keys
    SCROLL_LOCK = 'Scroll_Lock'
    NUM_LOCK = 'Num_Lock'
    CAPS_LOCK = 'Caps_Lock'

    # Navigation keys
    UP = 'Up'
    DOWN = 'Down'
    LEFT = 'Left'
    RIGHT = 'Right'
    PAGE_UP = 'Prior'
    PAGE_DOWN = 'Next'
    HOME = 'Home'
    END = 'End'

    # Number pad keys
    MULTIPLY = 'KP_Multiply'
    ADD = 'KP_Add'
    SEPARATOR = 'KP_Separator'
    SUBTRACT = 'KP_Subtract'
    DECIMAL = 'KP_Decimal'
    DIVIDE = 'KP_Divide'

    # Number pad number keys
    NUMPAD0 = 'KP_0'
    NUMPAD1 = 'KP_1'
    NUMPAD2 = 'KP_2'
    NUMPAD3 = 'KP_3'
    NUMPAD4 = 'KP_4'
    NUMPAD5 = 'KP_5'
    NUMPAD6 = 'KP_6'
    NUMPAD7 = 'KP_7'
    NUMPAD8 = 'KP_8'
    NUMPAD9 = 'KP_9'

    # Function keys
    F1 = 'F1'
    F2 = 'F2'
    F3 = 'F3'
    F4 = 'F4'
    F5 = 'F5'
    F6 = 'F6'
    F7 = 'F7'
    F8 = 'F8'
    F9 = 'F9'
    F10 = 'F10'
    F11 = 'F11'
    F12 = 'F12'
    F13 = 'F13'
    F14 = 'F14'
    F15 = 'F15'
    F16 = 'F16'
    F17 = 'F17'
    F18 = 'F18'
    F19 = 'F19'
    F20 = 'F20'
    F21 = 'F21'
    F22 = 'F22'
    F23 = 'F23'
    F24 = 'F24'

    # Multimedia keys
    VOLUME_DOWN = 'XF86AudioLowerVolume'
    VOLUME_MUTE = 'XF86AudioMute'
    VOLUME_UP = 'XF86AudioRaiseVolume'
    MEDIA_PREV_TRACK = 'XF86AudioPrev'
    MEDIA_NEXT_TRACK = 'XF86AudioNext'
    MEDIA_PLAY_PAUSE = 'XF86AudioPlay'
    BROWSER_BACK = 'XF86Back'
    BROWSER_FORWARD = 'XF86Forward'



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
BringApp and StartApp actions
============================================================================

The :class:`StartApp` and :class:`BringApp` action classes are used to
start an application and bring it to the foreground.  :class:`StartApp`
starts an application by running an executable file, while
:class:`BringApp` first checks whether the application is already running
and if so brings it to the foreground, otherwise starts it by running the
executable file.


Example usage
----------------------------------------------------------------------------

The following example brings Notepad to the foreground if it is already
open, otherwise it starts Notepad::

   BringApp(r"C:\\Windows\\system32\\notepad.exe").execute()

Note that the path to *notepad.exe* given above might not be correct for
your computer, since it depends on the operating system and its
configuration.

In some cases an application might be accessible simply through the file
name of its executable, without specifying the directory.  This depends on
the operating system's path configuration.  For example, on the author's
computer the following command successfully starts Notepad::

   BringApp("notepad").execute()

Applications on MacOS are started and switched to using the application
name::

   BringApp("System Preferences").execute()


Class reference
----------------------------------------------------------------------------

"""

import locale
import os.path
from subprocess           import Popen
import sys
import time

import six

from .action_base         import ActionBase, ActionError
from .action_focuswindow  import FocusWindow
from .action_waitwindow   import WaitWindow
from ..windows            import Window


#---------------------------------------------------------------------------

class StartApp(ActionBase):
    """
        Start an application.

        When this action is executed, it runs a file (executable),
        optionally with commandline arguments.

    """

    def __init__(self, *args, **kwargs):
        """
            Constructor arguments:
             - *args* (variable argument list of *str*'s) --
               these strings are passed to subprocess.Popen()
               to start the application as a child process
             - *cwd* (*str*, default *None*) --
               if not *None*, then start the application in this
               directory
             - *focus_after_start* (*bool*, default *False*) --
               if *True*, then attempt to bring the window to the foreground
               after starting the application.

            A single *list* or *tuple* argument can be used instead of
            variable arguments.

        """
        ActionBase.__init__(self)
        if len(args) == 1 and isinstance(args[0], (tuple, list)):
            args = args[0]  # use the sub-list instead

        self._args = args

        self._cwd = kwargs.pop("cwd", None)
        self._focus_after_start = kwargs.pop("focus_after_start", False)
        if kwargs:
            raise ActionError("Invalid keyword arguments: %r" % kwargs)

        # Expand any variables within path names.
        self._args = [self._interpret(a) for a in self._args]
        if self._cwd:
            self._cwd = self._interpret(self._cwd)

        self._str = str(", ".join(repr(a) for a in self._args))

    @classmethod
    def _interpret(cls, path):
        if not isinstance(path, six.string_types):
            raise TypeError("expected string argument for path, but got "
                            "%s" % path)

        # Encode text strings if this is Python 2.
        if six.PY2 and isinstance(path, six.text_type):
            encoding = locale.getpreferredencoding()
            path = path.encode(encoding)

        return os.path.expanduser(os.path.expandvars(path))

    def _start_app_process(self):
        try:
            return Popen(self._args, cwd=self._cwd)
        except Exception as e:
            raise ActionError("Failed to start app %s: %s" % (self._str, e))

    def _focus_window_after_starting(self, pid):
        timeout = 1.0
        exe = self._args[0]

        # Wait for the window to appear for N seconds.
        action = WaitWindow(executable=exe, timeout=timeout)
        if action.execute():
            # Bring the window to the foreground.
            Window.get_foreground().set_foreground()

        # The application window wasn't focused, so try to focus it.
        else:
            target = pid
            start = time.time()
            while time.time() - start < timeout:
                found = False
                for window in Window.get_matching_windows(exe):
                    if target is None or window.pid == target:
                        window.set_foreground()
                        found = True
                        break
                if found:
                    break

    def _darwin_start_app(self):
        # Try to use the macOS 'open' command-line program to start a new
        # instance of the specified application. Return False if this is
        # either not applicable or doesn't work.
        try_using_open = (
            sys.platform == "darwin" and len(self._args) == 1 and
            not os.path.isabs(self._args[0])
        )
        if try_using_open:
            process = Popen(['open', '-n', '-a', self._args[0]])
            return process.wait() == 0
        return False

    def _execute(self, data=None):
        self._log.debug("Starting app: %r", self._args)
        if self._darwin_start_app():
            pid = None
        else:
            # This either isn't macOS or the 'open' program didn't work, so
            # start the application as normal and get the process ID.
            pid = self._start_app_process().pid

        # If specified, focus the application after starting it.
        if self._focus_after_start:
            self._focus_window_after_starting(pid)


#---------------------------------------------------------------------------

class BringApp(StartApp):
    """
        Bring an application to the foreground, starting it if it is not
        yet running.

        When this action is executed, it looks for an existing window of
        the application specified in the constructor arguments.  If an
        existing window is found, that window is brought to the
        foreground.  On the other hand, if no window is found the
        application is started.

        Note that the constructor arguments are identical to those used by
        the :class:`StartApp` action class.

    """

    def __init__(self, *args, **kwargs):
        """
            Constructor arguments:
             - *args* (variable argument list of *str*'s) --
               these strings are passed to :meth:`subprocess.Popen`
               to start the application as a child process
             - *cwd* (*str*, default *None*) --
               if not *None*, then start the application in this
               directory
             - *title* (*str*, default *None*) --
               if not *None*, then match existing windows using this
               title.
             - *index* (*str* or *int*) -- zero-based index of the target
               window, for multiple matching windows; can be a string (for
               substitution) but must be convertible to an integer.
             - *filter_func* (*callable*) -- called with a single argument
               (the window object), and should return ``True`` for your
               target windows; example:
               ``lambda window: window.get_position().dy > 100``.
             - *focus_after_start* (*bool*, default *False*) --
               if *True*, then attempt to bring the window to the foreground
               after starting the application. Does nothing if the
               application is already running.
             - *focus_only* (*bool*, default *False*) -- if *True*, then
               attempt to focus a matching window without raising it by
               using the *set_focus()* method instead of *set_foreground()*.
               This argument may do nothing depending on the platform.

        """
        self._title = kwargs.pop("title", None)
        self._index = kwargs.pop("index", None)
        self._filter_func = kwargs.pop("filter_func", None)
        self._focus_only = kwargs.pop("focus_only", False)
        StartApp.__init__(self, *args, **kwargs)

    def _execute(self, data=None):
        self._log.debug("Bringing app: %r", self._args)
        target = self._args[0].lower()
        title = self._title
        index = self._index
        filter_func = self._filter_func
        focus_only = self._focus_only
        focus_action = FocusWindow(executable=target, title=title,
                                   index=index, filter_func=filter_func,
                                   focus_only=focus_only)
        # Attempt to focus on an existing window.
        if not focus_action.execute():
            # Failed to focus on an existing window, so start
            #  the application.
            StartApp._execute(self, data)



# This file is part of Aenea
#
# Aenea is free software: you can redistribute it and/or modify it under
# the terms of version 3 of the GNU Lesser General Public License as
# published by the Free Software Foundation.
#
# Aenea is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with Aenea.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright (2014) Alex Roper
# Alex Roper <alex@aroper.net>


'''
ContextAction
============================================================================

'''


from .action_base import ActionBase
from ..grammar.context import Context
from ..windows import Window


def _ensure_execution_context(data):
    '''Populates the data field of execute with context information if
    not present.'''
    if data is None:
        data = {}
    if '_context' not in data:
        data['_context'] = Window.get_foreground()
    return data

class ContextAction(ActionBase):
    '''
    Action class to execute a different action depending on which context is
    currently active.

    This is especially useful for allowing the same commands to work in
    multiple applications without redefining them in other grammars. An
    example of this is the redo shortcut. Some applications use
    *Ctrl+Shift+Z*, while others might use *Ctrl+Y* instead.
    ``ContextAction`` could be used to define *Ctrl+Y* as the default and
    use *Ctrl+Shift+Z* for specific contexts::

        redo = ContextAction(default=Key('c-y'), actions=[
            # Use cs-z for rstudio
            (AppContext(executable="rstudio"), Key('cs-z')),
        ])

    This class was originally written for the
    `Aenea <https://github.com/dictation-toolbox/aenea>`__ project by Alex
    Roper and has been modified to work without Aenea's functionality.
    '''
    def __init__(self, default=None, actions=None):
        '''
            Constructor arguments:
             - *default* (action object, default *do nothing*) -- the
               default action to execute if there was no matching context in
               *actions*.
             - *actions* (iterable, default *empty list*) -- an iterable
               object containing context-action pairs. The action of the
               first matching context will be executed.

        '''
        if actions is None:
            actions = []

        # Validate default action.
        # Use a new ActionBase action (to do nothing) if default is None.
        if default is None:
            self._log_init.debug("Using default action for ContextAction")
            default = ActionBase()

        # Otherwise check if default is an action.
        elif not isinstance(default, ActionBase):
            raise TypeError("Default action for ContextAction should be an "
                            "ActionBase or None, not %s." % default)

        # Set the default action.
        self.default = default

        # Validate the actions list.
        # Check if it can be converted to a dictionary. or convert it to a
        # is one already (for consistent evaluation order).
        if not isinstance(actions, dict):
            dict(actions)  # can raise a ValueError

        # Convert dictionaries to lists instead for consistent evaluation
        # order.
        else:
            actions = list(actions.items())

        # Check the types of all keys and values.
        for (context, action) in actions:
            if not isinstance(context, Context):
                raise TypeError("ContextAction actions list contains "
                                "unexpected object: %s, instead of a "
                                "Context." % context)
            if not isinstance(action, ActionBase):
                raise TypeError("ContextAction actions list contains "
                                "unexpected object: %s, instead of an "
                                "ActionBase." % action)

        # Set the actions list.
        self.actions = actions
        ActionBase.__init__(self)

    def add_context(self, context, action):
        '''
        Add a context-action pair to the *actions* list.

        :param context: dragonfly context
        :param action: dragonfly action
        :type context: Context
        :type action: ActionBase
        '''
        self.actions.append((context, action))

    def _execute(self, data=None):
        data = _ensure_execution_context(data)
        win = data['_context']
        try:
            for (context, action) in self.actions:
                if context.matches(win.executable, win.title, win.handle):
                    return action.execute(data)

            # Execute the default action and return the success.
            return self.default.execute(data)
        except Exception as e:
            self._log.exception("Exception from matching context or "
                                "executing action %s: %s", self._str, e)
            return False



#
# This file is part of Dragonfly.
# (c) Copyright 2019 by David Zurow
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Utility & testing classes for Kaldi backend
"""

import time
from contextlib import contextmanager

debug_timer_enabled = True
_debug_timer_stack = []

@contextmanager
def debug_timer(log, desc, enabled=True, independent=False):
    start_time = time.time()
    if not independent: _debug_timer_stack.append(start_time)
    spent_time_func = lambda: time.time() - start_time
    yield spent_time_func
    start_time_adjusted = _debug_timer_stack.pop() if not independent else 0
    if enabled:
        if debug_timer_enabled:
            log("%s %d ms" % (desc, (time.time() - start_time_adjusted) * 1000))
        if _debug_timer_stack and not independent:
            _debug_timer_stack[-1] += spent_time_func()



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
    This file implements a basic multiplexing interface to the current
    engine's timer.
"""


import time
import logging

from dragonfly.engines import get_engine


class _Timer(object):
    """
    This class is **deprecated** and will be removed in a future release.

    Please use the :meth:`engine.create_timer` method instead.

    """

    _log = logging.getLogger("timer")

    class Callback(object):
        def __init__(self, function, interval):
            self.function = function
            self.interval = interval
            self.next_time = time.time() + self.interval
            self.timer = None

        def call(self):
            self.next_time += self.interval
            try:
                self.function()
            except Exception as e:
                _Timer._log.exception("Exception during timer callback")
                print("Exception during timer callback: %s (%r)" % (e, e))

    def __init__(self, interval):
        self.interval = interval
        self.callbacks = []
        self._log.warning("Dragonfly's _Timer class has been deprecated. "
                          "Please use engine.create_timer() instead.")

    def add_callback(self, function, interval):
        t = get_engine().create_timer(function, interval)
        c = self.Callback(function, interval)
        c.timer = t
        self.callbacks.append(c)

    def remove_callback(self, function):
        for c in self.callbacks:
            if c.function == function:
                self.callbacks.remove(c)
                c.timer.stop()

    def callback(self):
        now = time.time()
        for c in self.callbacks:
            if c.next_time < now:
                c.timer.call()


timer = _Timer(0.025)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
MappingRule class
============================================================================

The MappingRule class is designed to make it very easy to create a rule
based on a mapping of spoken-forms to semantic values.

This class has the following parameters to customize its behavior:

 - *mapping* -- mapping of spoken-forms to semantic values
 - *extras* -- extras elements referenced from the compound spec
 - *defaults* -- default values for the extras
 - *exported* -- whether the rule is exported
 - *context* -- context in which the rule will be active

Each of these parameters can be passed as a (keyword) arguments to the
constructor, or defined as a class attribute in a derived class.


Example usage
............................................................................

The MappingRule class can be used to define a voice-command as follows::

    class ExampleRule(MappingRule):

        mapping  = {
                    "[feed] address [bar]":                Key("a-d"),
                    "subscribe [[to] [this] feed]":        Key("a-u"),
                    "paste [feed] address":                Key("a-d, c-v, enter"),
                    "feeds | feed (list | window | win)":  Key("a-d, tab:2, s-tab"),
                    "down [<n>] (feed | feeds)":           Key("a-d, tab:2, s-tab, down:%(n)d"),
                    "up [<n>] (feed | feeds)":             Key("a-d, tab:2, s-tab, up:%(n)d"),
                    "open [item]":                         Key("a-d, tab:2, c-s"),
                    "newer [<n>]":                         Key("a-d, tab:2, up:%(n)d"),
                    "older [<n>]":                         Key("a-d, tab:2, down:%(n)d"),
                    "mark all [as] read":                  Key("cs-r"),
                    "mark all [as] unread":                Key("cs-u"),
                    "search [bar]":                        Key("a-s"),
                    "search [for] <text>":                 Key("a-s") + Text("%(text)s\\n"),
                   }
        extras   = [
                    Integer("n", 1, 20),
                    Dictation("text"),
                   ]
        defaults = {
                    "n": 1,
                   }

    rule = ExampleRule()
    grammar.add_rule(rule)


Class reference
............................................................................

"""
from six import string_types

from .rule_base         import Rule
from .elements          import ElementBase, Compound, Alternative
from ..actions.actions  import ActionBase


#---------------------------------------------------------------------------

class MappingRule(Rule):
    """
        Rule class based on a mapping of spoken-forms to semantic values.

        Constructor arguments:
         - *name* (*str*) -- the rule's name
         - *mapping* (*dict*) -- mapping of spoken-forms to semantic
           values
         - *extras* (sequence) -- extras elements referenced from the
           spoken-forms in *mapping*
         - *defaults* (*dict*) -- default values for the extras
         - *exported* (boolean) -- whether the rule is exported
         - *context* (*Context*) -- context in which the rule will be active

    """

    mapping  = {}
    extras   = []
    defaults = {}
    context  = None
    _default_exported = True

    #-----------------------------------------------------------------------

    def __init__(self, name=None, mapping=None, extras=None, defaults=None,
                 exported=None, context=None):
        # pylint: disable=too-many-branches

        if name     is None: name     = self.__class__.__name__
        if mapping  is None: mapping  = self.mapping
        if extras   is None: extras   = self.extras
        if defaults is None: defaults = self.defaults
        if context  is None: context  = self.context

        # Complex handling of exported, because of clashing use of the
        #  exported name at the class level: property & class-value.
        if exported is not None:
            pass
        elif (hasattr(self.__class__, "exported")
              and not isinstance(self.__class__.exported, property)):
            exported = self.__class__.exported
        else:
            exported = self._default_exported

        # Type checking of initialization values.
        assert isinstance(name, string_types)
        assert isinstance(mapping, dict)
        for key, value in mapping.items():
            assert isinstance(key, string_types)
        assert isinstance(extras, (list, tuple))
        for item in extras:
            assert isinstance(item, ElementBase)
        assert exported in (True, False)

        self._name     = name
        self._mapping  = mapping
        self._extras   = {element.name : element for element in extras}
        self._defaults = defaults

        children = []
        for spec, value in self._mapping.items():
            c = Compound(spec, elements=self._extras, value=value)
            children.append(c)

        if children:  element = Alternative(children)
        else:         element = None
        Rule.__init__(self, self._name, element, exported=exported,
                      context=context)

    #-----------------------------------------------------------------------

    @property
    def specs(self):
        """
        Each spoken-form in the rule.
        :rtype: list
        """
        return [k for k, _ in self._mapping.items()]

    def value(self, node):
        node = node.children[0]
        value = node.value()

        if hasattr(value, "copy_bind"):
            # Prepare *extras* dict for passing to _copy_bind().
            extras = {
                "_grammar":  self.grammar,
                "_rule":     self,
                "_node":     node,
            }
            extras.update(self._defaults)
            for name, element in self._extras.items():
                extra_node = node.get_child_by_name(name, shallow=True)
                if extra_node:
                    extras[name] = extra_node.value()
                elif element.has_default():
                    extras[name] = element.default

            value = value.copy_bind(extras)

        return value

    def process_recognition(self, node):
        """
            Process a recognition of this rule.

            This method is called by the containing Grammar when
            this rule is recognized.  This method collects information
            about the recognition and then calls
            MappingRule._process_recognition.

            - *node* -- The root node of the recognition parse tree.
        """
        item_value = node.value()

        # Prepare *extras* dict for passing to _process_recognition().
        extras = {
            "_grammar":  self.grammar,
            "_rule":     self,
            "_node":     node,
        }
        extras.update(self._defaults)
        for name, element in self._extras.items():
            extra_node = node.get_child_by_name(name, shallow=True)
            if extra_node:
                extras[name] = extra_node.value()
            elif element.has_default():
                extras[name] = element.default

        # Call the method to do the actual processing.
        self._process_recognition(item_value, extras)

    def _process_recognition(self, value, extras):
        """
            Default recognition processing.

            This is the method which should be overridden in most cases
            to provide derived classes with custom recognition
            processing functionality.

            This default processing method takes the mapping value
            from the recognition and, if it is an action, executes it
            with the given extras as a dynamic values.

            - *value* -- The mapping value recognized.
            - *extras* -- A dictionary of all elements from the
              extras list contained within this recognition.
              Maps element name -> element value.
        """
        if isinstance(value, ActionBase):
            value.execute(extras)
        elif self._log_proc:
            self._log_proc.warning("%s: mapping value is not an action,"
                                   " cannot execute.", self)



#
# This file is part of Dragonfly.
# (c) Copyright 2021 by Dane Finlay
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
    This module demonstrates Dragonfly's window movement functionality.  A
    number of classes are employed to demo a good number of use cases with a
    smaller amount of code.

    This module can be activated in the same way as other Natlink macros
    by placing it in the "NatLink/MacroSystem" folder or in a folder loaded
    by one of Dragonfly's module loaders.

"""

# pylint: disable=invalid-name

from dragonfly import Grammar, CompoundRule, Compound, Window
from dragonfly.actions.mouse import  get_cursor_position


#---------------------------------------------------------------------------
# Create this module's grammar.

grammar = Grammar("Window move example")


#---------------------------------------------------------------------------
# Define classes for window movement.

class WindowMovementBase:
    """
        Base class for moving a target window.
    """

    def __init__(self):
        self.target_window = None

    def target_foreground_window(self):
        """
            Set the target window to the current foreground window.
        """
        self.target_window = Window.get_foreground()

    def move_target_window(self, window_animation):
        """
            Move the window, animating the movement of the window if
            specified.
        """
        target_window = self.target_window
        src_pos = target_window.get_position()
        dst_pos = self.calculate_target_window_destination(src_pos)
        target_window.move(dst_pos, window_animation)

    def calculate_target_window_destination(self, src_position):
        """
            Calculate and return the target window destination.

            This method should be overridden to return a Rectangle
            object representing where the window should be moved.
        """
        raise NotImplementedError()


class MoveWindowHere(WindowMovementBase):
    """
        Class for moving a target window to the current cursor position.
    """

    def calculate_target_window_destination(self, src_position):
        # The destination of the target window is the source position with
        #  its top left-hand corner at the current cursor position.
        x_mouse, y_mouse = get_cursor_position()
        src_position.x1 = x_mouse
        src_position.y1 = y_mouse
        return src_position


class CenterWindowHere(WindowMovementBase):
    """
        Class for centering a target window on the current cursor position.
    """

    def calculate_target_window_destination(self, src_position):
        # The destination of the target window is the source position
        #  centered on the current cursor position.
        x_mouse, y_mouse = get_cursor_position()
        relative_x_center = src_position.x_center - src_position.x1
        relative_y_center = src_position.y_center - src_position.y1
        src_position.x1 = x_mouse - relative_x_center
        src_position.y1 = y_mouse - relative_y_center
        return src_position


#---------------------------------------------------------------------------
# Define rule classes for window movement rules.

class MoveWindowRule(CompoundRule):
    """
        Base class for window movement rules.
    """

    #: The extra to use for optionally animating window movement.
    window_animation_extra_name = "animation"

    #-----------------------------------------------------------------------

    def __init__(self, window_movement, name=None, spec=None, extras=None,
                 defaults=None, exported=None, context=None):
        # pylint: disable=too-many-arguments
        CompoundRule.__init__(self, name, spec, extras, defaults, exported,
                              context)
        self.window_movement = window_movement

    #-----------------------------------------------------------------------

    def move_target_window(self, extras):
        """
            Move the target window.
        """
        # Get the value for the window animation extra (or None) and move
        #  the target window, passing the value.
        window_animation = extras.get(self.window_animation_extra_name)
        self.window_movement.move_target_window(window_animation)


class MoveForegroundWindowRule(MoveWindowRule):
    """
        Class for foreground window movement rules.
    """

    def _process_recognition(self, node, extras):
        # Set the foreground window as the target window and move it.
        self.window_movement.target_foreground_window()
        self.move_target_window(extras)


class MoveNotepadWindowRule(MoveWindowRule):
    """
        Class for notepad window movement rules.
    """

    def _process_recognition(self, node, extras):
        # Find a notepad window, set it as the target window and move it.
        #  Display a warning if no notepad window was found.
        notepad_windows = Window.get_matching_windows(executable="notepad",
                                                      title="notepad")
        if not notepad_windows:
            print("Could not find a notepad window to move.")
            return

        self.window_movement.target_window = notepad_windows[0]
        self.move_target_window(extras)


#---------------------------------------------------------------------------
# Create window movement objects to be used by our window movement rules.
move_window_here = MoveWindowHere()
center_window_here = CenterWindowHere()

# Define common extras for movement rules: an <animation> extra for
#  optionally animating window movement.
common_extras = [
    Compound(
        spec="animated", name="animation",

        # Possible window animation values: "linear", "spline".
        value_func=lambda _, __: "spline"
    )
]

# Create a rule for moving the foreground window to the current cursor
#  position.
rule1 = MoveForegroundWindowRule(
    window_movement=move_window_here,
    name="rule1",
    spec="move [this] window here [<animation>]",
    extras=common_extras
)

# Create a rule for centering the foreground window on the current cursor
#  position.
rule2 = MoveForegroundWindowRule(
    window_movement=center_window_here,
    name="rule2",
    spec="center [this] window here [<animation>]",
    extras=common_extras
)

# Create a rule for moving a notepad window to the current cursor position.
rule3 = MoveNotepadWindowRule(
    window_movement=move_window_here,
    name="rule3",
    spec="move notepad [window] here [<animation>]",
    extras=common_extras
)

# Create a rule for centering a notepad window on the current cursor
#  position.
rule4 = MoveNotepadWindowRule(
    window_movement=center_window_here,
    name="rule4",
    spec="center notepad [window] here [<animation>]",
    extras=common_extras
)

# Add our rules to the grammar instance.
grammar.add_rule(rule1)
grammar.add_rule(rule2)
grammar.add_rule(rule3)
grammar.add_rule(rule4)


#---------------------------------------------------------------------------
# Load the grammar instance and define how to unload it.

grammar.load()

# Unload function which will be called by natlink at unload time.
def unload():
    global grammar
    if grammar:
        grammar.unload()
    grammar = None



#
# This file is part of Dragonfly.
# (c) Copyright 2018 by Dane Finlay
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#


import unittest

from dragonfly import CompoundRule, MimicFailure, Grammar, get_engine
from dragonfly.test import (RuleTestCase, TestContext, RuleTestGrammar)


# ==========================================================================

class TestRules(RuleTestCase):

    def setUp(self):
        RuleTestCase.setUp(self)
        engine = get_engine()
        if engine.name == 'natlink':
            # Stop Dragon from dictating text for the duration of these
            # tests. This is required when testing for mimic failures.
            self.temp_grammar = Grammar("temp")
            self.temp_grammar.add_rule(CompoundRule(spec="exclusive rule"))
            self.temp_grammar.load()
            self.temp_grammar.set_exclusiveness(True)

    def tearDown(self):
        RuleTestCase.tearDown(self)
        engine = get_engine()
        if engine.name == 'natlink':
            self.temp_grammar.set_exclusiveness(False)
            self.temp_grammar.unload()

    def process_grammars_context(self):
        engine = get_engine()
        if engine.name.startswith("sapi5"):
            engine.process_grammars_context()

    def test_multiple_rules(self):
        """ Verify that the engine successfully mimics each rule in a
            grammar with multiple rules. """
        self.add_rule(CompoundRule(name="r1", spec="hello"))
        self.add_rule(CompoundRule(name="r2", spec="see you"))
        assert self.recognize_node("hello").words() == ["hello"]
        assert self.recognize_node("see you").words() == ["see", "you"]

    def test_rule_context(self):
        """ Verify that the engine works correctly with rule contexts. """
        context = TestContext(True)
        self.add_rule(CompoundRule(name="r1", spec="test context",
                                   context=context))

        # Test that the rule matches when in-context.
        results = self.recognize_node("test context").words()
        assert results == ["test", "context"]

        # Go out of context and test again.
        # Use the engine's mimic method because recognize_node won't return
        # RecognitionFailure like ElementTester.recognize does.
        context.active = False
        self.process_grammars_context()
        try:
            self.grammar.load()
            self.grammar.set_exclusiveness(True)
            self.assertRaises(MimicFailure, self.engine.mimic, "test context")
        finally:
            self.grammar.set_exclusiveness(False)
            self.grammar.unload()

        # Test again after going back into context.
        context.active = True
        self.process_grammars_context()
        results = self.recognize_node("test context").words()
        assert results == ["test", "context"]

    def test_grammar_context(self):
        """ Verify that the engine works correctly with grammar
            contexts."""
        # Recreate the RuleTestGrammar using a context and add a rule.
        context = TestContext(True)
        self.grammar = RuleTestGrammar(context=context)
        self.add_rule(CompoundRule(name="r1", spec="test context"))
        self.grammar.load()

        # Test that the rule matches when in-context.
        results = self.recognize_node("test context").words()
        assert results == ["test", "context"]

        # Go out of context and test again.
        context.active = False
        self.process_grammars_context()
        try:
            self.grammar.set_exclusiveness(True)
            self.assertRaises(MimicFailure, self.engine.mimic, "test context")
        finally:
            self.grammar.set_exclusiveness(False)

        # Test again after going back into context.
        context.active = True
        self.process_grammars_context()
        results = self.recognize_node("test context").words()
        assert results == ["test", "context"]

    def test_exclusive_grammars(self):
        """ Verify that the engine supports exclusive grammars. """
        # This is here as grammar exclusivity is context related.
        # Set up two grammars to test with.
        class TestRule(CompoundRule):
            def __init__(self, spec):
                CompoundRule.__init__(self, spec=spec)
                self.words = None

            def _process_recognition(self, node, extras):
                self.words = self.spec

        grammar1 = Grammar(name="Grammar1")
        grammar1.add_rule(TestRule(spec="grammar one"))
        grammar2 = Grammar(name="Grammar2")
        grammar2.add_rule(TestRule(spec="grammar two"))
        grammar3 = Grammar(name="Grammar3")
        grammar3.add_rule(TestRule(spec="grammar three"))
        grammar1.load()
        grammar2.load()
        grammar3.load()

        # Set grammar1 as exclusive and make some assertions.
        grammar1.set_exclusiveness(True)
        self.engine.mimic("grammar one")
        assert grammar1.rules[0].words == "grammar one"
        self.assertRaises(MimicFailure, self.engine.mimic, "grammar two")

        # Set grammar2 as exclusive and make some assertions.
        # Both exclusive grammars should be active.
        grammar2.set_exclusiveness(True)
        self.engine.mimic("grammar one")
        assert grammar1.rules[0].words == "grammar one"
        self.engine.mimic("grammar two")
        assert grammar2.rules[0].words == "grammar two"

        # Non-exclusive grammar 'grammar3' should not be active.
        self.assertRaises(MimicFailure, self.engine.mimic, "grammar three")

        # Set both grammars as no longer exclusive and make some assertions.
        grammar1.set_exclusiveness(False)
        grammar2.set_exclusiveness(False)
        if get_engine().name == 'natlink':
            self.temp_grammar.set_exclusiveness(False)
        self.engine.mimic("grammar one")
        assert grammar1.rules[0].words == "grammar one"
        self.engine.mimic("grammar two")
        assert grammar2.rules[0].words == "grammar two"
        self.engine.mimic("grammar three")
        assert grammar3.rules[0].words == "grammar three"

# ==========================================================================

if __name__ == "__main__":
    unittest.main()



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

# pylint: disable=E0401
# This file imports a Win32-only module.

import ctypes
import logging

import win32api

from .base_monitor import BaseMonitor
from .rectangle import Rectangle

# Force the current process to be DPI aware. This will ensure that the
# resolution is reported properly even when DPI scaling is on. This appears to
# be set to PROCESS_SYSTEM_DPI_AWARE when running embedded in Dragon, which
# provides awareness of the DPI when the application starts, but not when it
# changes. Unfortunately, this cannot be changed (presumably it is declared in a
# manifest). When running from the command line, however, the default is
# PROCESS_DPI_UNAWARE and can be adjusted here.
try:
    log = logging.getLogger("monitor.init")
    value = 2  # PROCESS_PER_MONITOR_DPI_AWARE
    hresult = ctypes.windll.shcore.SetProcessDpiAwareness(value)
    if hresult == -2147024809:  # E_INVALIDARG
        log.warning("DPI awareness could not be set; "
                    "SetProcessDpiAwareness() received an invalid "
                    "argument: %d", value)
    elif hresult == -2147024891:  # E_ACCESSDENIED
        log.warning("DPI awareness could not be set; it has been set "
                    "already.")
except OSError:
    # Do nothing if SetProcessDpiAwareness() could not be called.
    pass


#===========================================================================
# Monitor class for storing info about a single display monitor.

class Win32Monitor(BaseMonitor):
    """
    The monitor class used on Win32.
    """

    #-----------------------------------------------------------------------
    # Class methods to create new Monitor objects.

    @classmethod
    def get_all_monitors(cls):
        # Get an updated list of monitors.
        monitors = []
        for py_handle, _, rectangle in win32api.EnumDisplayMonitors():
            # Get the monitor handle.
            handle = py_handle.handle

            # Create a rectangle object representing the monitor's
            # dimensions and relative position.
            top_left_x, top_left_y, bottom_right_x, bottom_right_y = rectangle
            dx = bottom_right_x - top_left_x
            dy = bottom_right_y - top_left_y
            rectangle = Rectangle(top_left_x, top_left_y, dx, dy)

            # Get a new or updated monitor object and add it to the list.
            # Ensure that the primary monitor is first on the list.
            monitor = cls.get_monitor(handle, rectangle)
            if monitor.is_primary:
                monitors.insert(0, monitor)
            else:
                monitors.append(monitor)

        return monitors

    #-----------------------------------------------------------------------
    # Methods that control attribute access.

    @property
    def is_primary(self):
        """
        Whether this is the primary display monitor.

        :rtype: bool
        :returns: true or false
        """
        monitor_info = win32api.GetMonitorInfo(self._handle)
        return monitor_info["Flags"] & 1 == 1

    @property
    def name(self):
        """
        The device name of this monitor.

        :rtype: str
        :returns: monitor name
        """
        monitor_info = win32api.GetMonitorInfo(self._handle)
        return monitor_info["Device"]



#
# This file is part of Dragonfly.
# (c) Copyright 2019 by David Zurow
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Kaldi engine classes
"""

import collections, logging, os, sys, time

from packaging.version import Version
from six import string_types, print_, reraise
from six.moves import zip
import kaldi_active_grammar
from kaldi_active_grammar       import KaldiError, KaldiRule

from ..base                     import (EngineBase,
                                        EngineError,
                                        MimicFailure,
                                        DelegateTimerManager,
                                        DelegateTimerManagerInterface,
                                        DictationContainerBase,
                                        GrammarWrapperBase)
from .audio                     import MicAudio, VADAudio, AudioStore, WavAudio
from .dictation                 import user_dictation_list, user_dictation_dictlist
from .recobs                    import KaldiRecObsManager
from .testing                   import debug_timer
from dragonfly.grammar.state    import State
from dragonfly.windows          import Window

# Import the Kaldi compiler class. Suppress metaclass TypeErrors raised
# during documentation builds caused by mocking KAG.
try:
    from .compiler                  import KaldiCompiler
except TypeError:
    if os.environ.get("SPHINX_BUILD_RUNNING"):
        KaldiCompiler = None
    else:
        reraise(*sys.exc_info())

nan = float('nan')

#===========================================================================

class KaldiEngine(EngineBase, DelegateTimerManagerInterface):
    """ Speech recognition engine back-end for Kaldi recognizer. """

    _name = "kaldi"
    DictationContainer = DictationContainerBase
    _required_kag_version = "2.1.0"

    #-----------------------------------------------------------------------

    def __init__(self, model_dir=None, tmp_dir=None, input_device_index=None,
        audio_input_device=None, audio_self_threaded=True, audio_auto_reconnect=True, audio_reconnect_callback=None,
        retain_dir=None, retain_audio=None, retain_metadata=None, retain_approval_func=None,
        vad_aggressiveness=3, vad_padding_start_ms=150, vad_padding_end_ms=200, vad_complex_padding_end_ms=600,
        auto_add_to_user_lexicon=False, lazy_compilation=True, invalidate_cache=False,
        expected_error_rate_threshold=None,
        alternative_dictation=None,
        compiler_init_config=None, decoder_init_config=None,
        ):
        EngineBase.__init__(self)
        DelegateTimerManagerInterface.__init__(self)

        try:
            import kaldi_active_grammar, sounddevice, webrtcvad
        except ImportError as e:
            self._log.error("%s: Failed to import Kaldi engine "
                            "dependencies: %s", self, e)
            raise EngineError("Failed to import Kaldi engine dependencies.")

        # Compatible release version specification
        # https://stackoverflow.com/questions/11887762/how-do-i-compare-version-numbers-in-python/21065570
        required_kag_version = Version(self._required_kag_version)
        kag_version = Version(kaldi_active_grammar.__version__)
        if not ((kag_version >= required_kag_version) and (kag_version.release[0:2] == required_kag_version.release[0:2])):
            self._log.error("%s: Incompatible kaldi_active_grammar version %s! Expected ~= %s!" % (self, kag_version, required_kag_version))
            self._log.error("See https://dragonfly2.readthedocs.io/en/latest/kaldi_engine.html#updating-to-a-new-version")
            if not os.environ.get('DRAGONFLY_DEVELOP'):
                raise EngineError("Incompatible kaldi_active_grammar version")

        # Handle engine parameters
        if input_device_index is not None:
            if audio_input_device is not None:
                raise ValueError("Cannot set both input_device_index and audio_input_device")
            self._log.warning("%s: input_device_index is deprecated; please use audio_input_device", self)
            audio_input_device = int(input_device_index)
        if audio_input_device not in (None, False) and not isinstance(audio_input_device, (int, string_types)):
            raise TypeError("Invalid audio_input_device not int or string: %r" % (audio_input_device,))
        if audio_reconnect_callback is not None and not callable(audio_reconnect_callback):
            raise TypeError("Invalid audio_reconnect_callback not callable: %r" % (audio_reconnect_callback,))
        if retain_dir is not None and not isinstance(retain_dir, string_types):
            raise TypeError("Invalid retain_dir not string: %r" % (retain_dir,))
        if retain_audio and not retain_dir:
            raise ValueError("retain_audio=True requires retain_dir to be set")
        if retain_approval_func is not None and not callable(retain_approval_func):
            raise TypeError("Invalid retain_approval_func not callable: %r" % (retain_approval_func,))

        self._options = dict(
            model_dir = model_dir,
            tmp_dir = tmp_dir,
            audio_input_device = audio_input_device,
            audio_self_threaded = bool(audio_self_threaded),
            audio_auto_reconnect = bool(audio_auto_reconnect),
            audio_reconnect_callback = audio_reconnect_callback,
            retain_dir = retain_dir,
            retain_audio = bool(retain_audio) if retain_audio is not None else bool(retain_dir),
            retain_metadata = bool(retain_metadata) if retain_metadata is not None else bool(retain_dir),
            retain_approval_func = retain_approval_func,
            vad_aggressiveness = int(vad_aggressiveness),
            vad_padding_start_ms = int(vad_padding_start_ms),
            vad_padding_end_ms = int(vad_padding_end_ms),
            vad_complex_padding_end_ms = int(vad_complex_padding_end_ms),
            auto_add_to_user_lexicon = bool(auto_add_to_user_lexicon),
            lazy_compilation = bool(lazy_compilation),
            invalidate_cache = bool(invalidate_cache),
            expected_error_rate_threshold = float(expected_error_rate_threshold) if expected_error_rate_threshold is not None else None,
            alternative_dictation = alternative_dictation,
            compiler_init_config = dict(compiler_init_config) if compiler_init_config else {},
            decoder_init_config = dict(decoder_init_config) if decoder_init_config else {},
        )

        # Setup
        self._reset_state()
        self._recognition_observer_manager = KaldiRecObsManager(self)
        self._timer_manager = DelegateTimerManager(0.02, self)

    def _reset_state(self):
        self._compiler = None
        self._decoder = None
        self._audio = None
        self._audio_iter = None
        self.audio_store = None

        self._loadunload_queue = collections.deque()
        self._grammar_wrappers_copy = {}
        self._any_exclusive_grammars = False
        self._saving_adaptation_state = False
        self._ignore_current_phrase = False
        self._in_phrase = False
        self._doing_recognition = False
        self._deferred_disconnect = False

    def connect(self):
        """ Connect to back-end SR engine. """
        if self._decoder:
            return
        self._reset_state()

        self._log.info("Loading Kaldi-Active-Grammar v%s in process %s." % (kaldi_active_grammar.__version__, os.getpid()))
        self._log.info("Kaldi options: %s" % self._options)

        self._compiler = KaldiCompiler(self._options['model_dir'], tmp_dir=self._options['tmp_dir'],
            auto_add_to_user_lexicon=self._options['auto_add_to_user_lexicon'],
            lazy_compilation=self._options['lazy_compilation'],
            alternative_dictation=self._options['alternative_dictation'],
            **self._options['compiler_init_config']
            )
        if self._options['invalidate_cache']:
            self._compiler.fst_cache.invalidate()

        self._decoder = self._compiler.init_decoder(config=self._options['decoder_init_config'])

        if self._options['audio_input_device'] is not False:
            self._audio = VADAudio(
                aggressiveness=self._options['vad_aggressiveness'],
                start=False,
                input_device=self._options['audio_input_device'],
                self_threaded=self._options['audio_self_threaded'],
                reconnect_callback=self._options['audio_reconnect_callback'],
                )
            self._audio_iter = self._audio.vad_collector(nowait=True,
                audio_auto_reconnect=self._options['audio_auto_reconnect'],
                start_window_ms=self._options['vad_padding_start_ms'],
                end_window_ms=self._options['vad_padding_end_ms'],
                complex_end_window_ms=self._options['vad_complex_padding_end_ms'],
                )
            self.audio_store = AudioStore(self._audio, maxlen=(1 if self._options['retain_dir'] else 0),
                save_dir=self._options['retain_dir'], save_audio=self._options['retain_audio'], save_metadata=self._options['retain_metadata'],
                retain_approval_func=self._options['retain_approval_func'])

    def disconnect(self):
        """ Disconnect from back-end SR engine. Exits from ``do_recognition()``. """
        if self._doing_recognition:
            self._deferred_disconnect = True
        else:
            if self._audio:
                self._audio.destroy()
            if self.audio_store:
                self.audio_store.save_all()
            self._reset_state()
            self._grammar_wrappers = {}  # From EngineBase

    @staticmethod
    def print_mic_list():
        MicAudio.print_list()

    def _apply_win32_kb_input_logging_fix(self):
        # Hack to avoid bug processing keyboard actions on Windows
        if os.name == 'nt':
            action_exec_logger = logging.getLogger('action.exec')
            if action_exec_logger.getEffectiveLevel() > logging.DEBUG:
                self._log.warning("%s: Enabling logging of actions "
                                  "execution to avoid bug processing "
                                  "keyboard actions on Windows", self)
                action_exec_logger.setLevel(logging.DEBUG)

    #-----------------------------------------------------------------------
    # Methods for working with grammars.

    def _load_grammar(self, grammar):
        """ Load the given *grammar*. """
        if not self._decoder:
            self.connect()

        self._log.info("Loading grammar %s" % grammar.name)
        kaldi_rule_by_rule_dict = self._compiler.compile_grammar(grammar, self)
        wrapper = GrammarWrapper(grammar, kaldi_rule_by_rule_dict, self,
                                 self._recognition_observer_manager)

        def load():
            for (rule, kaldi_rule) in kaldi_rule_by_rule_dict.items():
                kaldi_rule.active = bool(rule.active)  # Initialize to correct activity
                kaldi_rule.load(lazy=self._compiler.lazy_compilation)
        if self._in_phrase:
            self._loadunload_queue.append(load)
        else:
            load()

        return wrapper

    def _unload_grammar(self, grammar, wrapper):
        """ Unload the given *grammar*. """
        self._log.debug("Unloading grammar %s." % grammar.name)
        def unload():
            rules = list(wrapper.kaldi_rule_by_rule_dict.keys())
            self._compiler.unload_grammar(grammar, rules, self)
        if self._in_phrase:
            self._loadunload_queue.append(unload)
        else:
            unload()

    def activate_grammar(self, grammar):
        """ Activate the given *grammar*. """
        self._log.debug("Activating grammar %s." % grammar.name)
        self._get_grammar_wrapper(grammar).active = True

    def deactivate_grammar(self, grammar):
        """ Deactivate the given *grammar*. """
        self._log.debug("Deactivating grammar %s." % grammar.name)
        self._get_grammar_wrapper(grammar).active = False

    def activate_rule(self, rule, grammar):
        """ Activate the given *rule*. """
        self._log.debug("Activating rule %s in grammar %s." % (rule.name, grammar.name))
        self._compiler.kaldi_rule_by_rule_dict[rule].active = True

    def deactivate_rule(self, rule, grammar):
        """ Deactivate the given *rule*. """
        self._log.debug("Deactivating rule %s in grammar %s." % (rule.name, grammar.name))
        self._compiler.kaldi_rule_by_rule_dict[rule].active = False

    def update_list(self, lst, grammar):
        self._compiler.update_list(lst, grammar)

    def set_exclusiveness(self, grammar, exclusive):
        self._log.debug("Setting exclusiveness of grammar %s to %s." % (grammar.name, exclusive))
        self._get_grammar_wrapper(grammar).exclusive = exclusive
        if exclusive:
            self._get_grammar_wrapper(grammar).active = True
        self._any_exclusive_grammars = any(gw.exclusive for gw in self._grammar_wrappers.values())

    #-----------------------------------------------------------------------
    # Miscellaneous methods.

    def mimic(self, words):
        """ Mimic a recognition of the given *words*. """
        self._log.debug("Start of mimic: %s" % repr(words))
        try:
            output = words if isinstance(words, string_types) else " ".join(words)
            output = self._compiler.untranslate_output(output)
        except Exception as e:
            raise MimicFailure("Invalid mimic input %r: %s." % (words, e))

        self._recognition_observer_manager.notify_begin()
        kaldi_rules_activity = self._compute_kaldi_rules_activity()
        self.prepare_for_recognition()  # Redundant?

        recognition = self._parse_recognition(output, mimic=True)
        if not recognition.kaldi_rule:
            recognition.fail(mimic=True)
            raise MimicFailure("No matching rule found for %r." % (output,))
        recognition.process(mimic=True)
        self._log.debug("End of mimic: rule %s, %r" % (recognition.kaldi_rule, output))
        if not self._log.isEnabledFor(10):
            self._log.log(15, "End of mimic: rule %s, %r" % (recognition.kaldi_rule, output))

    def speak(self, text):
        """ Speak the given *text* using text-to-speech. """
        # FIXME
        self._log.warning("Text-to-speech is not implemented for this engine; printing text instead.")
        print_(text)

    def _get_language(self):
        return "en"

    def _has_quoted_words_support(self):
        return False

    def prepare_for_recognition(self):
        """ Can be called optionally before ``do_recognition()`` to speed up its starting of active recognition. """
        if self._in_phrase:
            self._log.warning("prepare_for_recognition ignored while in phrase; will be run after")
            return
        try:
            while self._loadunload_queue:
                operation = self._loadunload_queue.popleft()
                operation()
            self._compiler.prepare_for_recognition()
        except KaldiError as e:
            if len(e.args) >= 2 and isinstance(e.args[1], KaldiRule):
                kaldi_rule = e.args[1]
                raise self._compiler.make_compiler_error_for_kaldi_rule(kaldi_rule)
            raise

    def _do_recognition(self, timeout=None, single=False, audio_iter=None):
        """
            Loops performing recognition, by default forever, or for *timeout* seconds, or for a single recognition if *single=True*.
            Returns ``False`` if timeout occurred without a recognition.
        """
        self._log.debug("do_recognition: timeout %s" % timeout)
        if not self._decoder:
            raise EngineError("Cannot recognize before connect()")
        if audio_iter is None and self._audio is None:
            raise EngineError("No audio input")
        self._doing_recognition = True
        self._in_phrase = False
        self._ignore_current_phrase = False
        in_complex = False
        end_time = None
        timed_out = False

        try:
            self.prepare_for_recognition()  # Try to get compilation out of the way before starting audio

            if timeout != None:
                end_time = time.time() + timeout
                timed_out = True

            if audio_iter == None:
                self._audio.start()
                audio_iter = self._audio_iter
            self._log.info("Listening...")
            next(audio_iter)  # Prime the audio iterator

            # Loop until timeout (if set) or until disconnect() is called.
            while (not self._deferred_disconnect) and ((not end_time) or (time.time() < end_time)):
                block = audio_iter.send(in_complex)

                if block is False:
                    # No audio block available
                    time.sleep(0.001)

                elif block is not None:
                    if not self._in_phrase:
                        # Start of phrase
                        self._recognition_observer_manager.notify_begin()
                        with debug_timer(self._log.debug, "computing activity"):
                            kaldi_rules_activity = self._compute_kaldi_rules_activity()
                        self._in_phrase = True
                        self._ignore_current_phrase = False
                        self._grammar_wrappers_copy = self._grammar_wrappers.copy()  # Keep a copy of valid grammar wrappers as of the start of utterance

                    else:
                        # Ongoing phrase
                        kaldi_rules_activity = None
                    self._decoder.decode(block, False, kaldi_rules_activity)
                    if self.audio_store:
                        self.audio_store.add_block(block)
                    output, info = self._decoder.get_output()
                    self._log.log(5, "Partial phrase: %r [in_complex=%s]", output, in_complex)
                    kaldi_rule, words, words_are_dictation_mask, in_dictation = self._compiler.parse_partial_output(output)
                    in_complex = bool(in_dictation or (kaldi_rule and kaldi_rule.is_complex))

                else:
                    # End of phrase
                    self._decoder.decode(b'', True)
                    output, info = self._decoder.get_output()
                    if not self._ignore_current_phrase:
                        expected_error_rate = info.get('expected_error_rate', nan)
                        confidence = info.get('confidence', nan)
                        # output = self._compiler.untranslate_output(output)
                        recognition = self._parse_recognition(output)
                        is_acceptable_recognition = recognition.kaldi_rule and (recognition.has_dictation or not (
                            self._options['expected_error_rate_threshold'] and (expected_error_rate > self._options['expected_error_rate_threshold'])
                        ))
                        if is_acceptable_recognition:
                            recognition.process(expected_error_rate=expected_error_rate, confidence=confidence)
                        else:
                            recognition.fail(expected_error_rate=expected_error_rate, confidence=confidence)

                        kaldi_rule, parsed_output = recognition.kaldi_rule, recognition.parsed_output
                        self._log.log(15, "End of phrase: eer=%.2f conf=%.2f%s, rule %s, %r",
                            expected_error_rate, confidence, (" [BAD]" if not is_acceptable_recognition else ""), kaldi_rule, parsed_output)
                        if self._saving_adaptation_state and is_acceptable_recognition:  # Don't save adaptation state for bad recognitions
                            self._decoder.save_adaptation_state()
                        if self.audio_store:
                            if kaldi_rule and is_acceptable_recognition:  # Don't store audio/metadata for bad recognitions
                                self.audio_store.finalize(parsed_output,
                                    kaldi_rule.parent_grammar.name, kaldi_rule.parent_rule.name,
                                    likelihood=expected_error_rate, has_dictation=kaldi_rule.has_dictation)
                            else:
                                self.audio_store.cancel()

                    self._in_phrase = False
                    self._ignore_current_phrase = False
                    in_complex = False
                    timed_out = False
                    if single:
                        break
                    self.prepare_for_recognition()  # Do any of this leftover, now that phrase is done

                self.call_timer_callback()

        except StopIteration:
            if audio_iter == self._audio_iter:
                self._log.warning("audio iterator stopped unexpectedly")

        finally:
            self._doing_recognition = False
            if (audio_iter == self._audio_iter) and self._audio:
                try:
                    self._audio.stop()  # We started the audio above, so we should stop it
                except Exception:
                    self._log.exception("Error stopping audio")
            if self._deferred_disconnect:
                self.disconnect()

        return not timed_out

    in_phrase = property(lambda self: self._in_phrase,
        doc="Whether or not the engine is currently in the middle of hearing a phrase from the user.")

    def recognize_wave_file(self, filename, realtime=False, **kwargs):
        """
            Does recognition on given wave file, treating it as a single
            utterance (without VAD), then returns.
        """
        self.do_recognition(audio_iter=WavAudio.read_file(filename, realtime=realtime), **kwargs)

    def recognize_wave_file_as_stream(self, filename, realtime=False, **kwargs):
        """
            Does recognition on given wave file, treating it as a stream and
            processing it with VAD to break it into multiple utterances (as with
            normal microphone audio input), then returns.
        """
        self.do_recognition(audio_iter=WavAudio.read_file_with_vad(filename, realtime=realtime), **kwargs)

    def ignore_current_phrase(self):
        """
            Marks the current phrase's recognition to be ignored when it completes, or does nothing if there is none.
            Returns *bool* indicating whether or not there was a current phrase being heard.
        """
        if not self.in_phrase:
            return False
        self._ignore_current_phrase = True
        return True

    saving_adaptation_state = property(lambda self: self._saving_adaptation_state,
        doc="Whether or not the engine is currently automatically saving adaptation state between utterances.")
    @saving_adaptation_state.setter
    def saving_adaptation_state(self, value):
        self._saving_adaptation_state = value

    def start_saving_adaptation_state(self):
        """ Enable automatic saving of adaptation state between utterances, which may improve recognition accuracy in the short term, but is not stored between runs. """
        self.saving_adaptation_state = True

    def stop_saving_adaptation_state(self):
        """ Disables automatic saving of adaptation state between utterances, which you might want to do when you expect there to be noise and don't want it to pollute your current adaptation state. """
        self.saving_adaptation_state = False

    def reset_adaptation_state(self):
        self._decoder.reset_adaptation_state()

    def add_word_list_to_user_dictation(self, word_list):
        """ Make UserDictation elements able to recognize each item of given
            list of strings *word_list*. Note: all characters will be converted
            to lowercase, and recognized as such. """
        word_list = [str(word).lower() for word in word_list]
        word_list = [word for word in word_list
            if word not in user_dictation_list]
        # if any((word != word.lower()) for word in word_list):
        #     raise ValueError("Cannot recognize words with uppercase")
        user_dictation_list.extend(word_list)

    def add_word_dict_to_user_dictation(self, word_dict):
        """ Make UserDictation elements able to recognize each item of given
            dict of strings *word_dict*. The key is the "spoken form" (which is
            recognized), and the value is the "written form" (which is returned
            as the text in the UserDictation element). Note: all characters in
            the keys will be converted to lowercase, but the values are returned
            as text verbatim. """
        word_dict = {str(key).lower(): str(value) for key, value in word_dict.items()}
        word_dict = {key: value for key, value in word_dict.items()
            if key not in user_dictation_dictlist.keys()}
        # if any((word != word.lower()) for word in word_dict.keys()):
        #     raise ValueError("Cannot recognize words with uppercase")
        user_dictation_dictlist.update(word_dict)

    #-----------------------------------------------------------------------
    # Internal processing methods.

    def _iter_all_grammar_wrappers_dynamically(self):
        """ Accounts for grammar wrappers being dynamically added during iteration. """
        processed_grammar_wrappers = set()
        todo_grammar_wrappers = set(self._grammar_wrappers.values())
        while todo_grammar_wrappers:
            while todo_grammar_wrappers:
                grammar_wrapper = todo_grammar_wrappers.pop()
                yield grammar_wrapper
                processed_grammar_wrappers.add(grammar_wrapper)
            todo_grammar_wrappers = set(self._grammar_wrappers.values()) - processed_grammar_wrappers

    def _compute_kaldi_rules_activity(self, phrase_start=True):
        window_info = {}
        if phrase_start:
            fg_window = Window.get_foreground()
            window_info = {
                "executable": fg_window.executable,
                "title": fg_window.title,
                "handle": fg_window.handle,
            }
            for grammar_wrapper in self._iter_all_grammar_wrappers_dynamically():
                grammar_wrapper.phrase_start_callback(**window_info)
        self.prepare_for_recognition()
        self._active_kaldi_rules = set()
        self._kaldi_rules_activity = [False] * self._compiler.num_kaldi_rules
        for grammar_wrapper in self._iter_all_grammar_wrappers_dynamically():
            if grammar_wrapper.active and (not self._any_exclusive_grammars or grammar_wrapper.exclusive):
                for kaldi_rule in grammar_wrapper.kaldi_rule_by_rule_dict.values():
                    if kaldi_rule.active:
                        self._active_kaldi_rules.add(kaldi_rule)
                        self._kaldi_rules_activity[kaldi_rule.id] = True
        self._log.debug("active kaldi_rules (from window %s): %s", window_info, [kr.name for kr in self._active_kaldi_rules])
        return self._kaldi_rules_activity

    def _parse_recognition(self, output, mimic=False):
        if mimic or self._compiler.parsing_framework == 'text':
            with debug_timer(self._log.debug, "kaldi_rule parse time"):
                detect_ambiguity = False
                results = []
                for kaldi_rule in sorted(self._active_kaldi_rules, key=lambda kr: 100 if kr.has_dictation else 0):
                    self._log.debug("attempting to parse %r with %s", output, kaldi_rule)
                    words = self._compiler.parse_output_for_rule(kaldi_rule, output)
                    if words is None:
                        continue
                    # self._log.debug("success %d", kaldi_rule_id)
                    # Pass (kaldi_rule, words) to below.
                    results.append((kaldi_rule, words))
                    if not detect_ambiguity:
                        break

                if not results:
                    if not mimic:
                        # We should never receive an unparsable recognition from kaldi, only from mimic
                        self._log.error("unable to parse recognition: %r" % output)
                    return Recognition.construct_empty(self)

                if len(results) > 1:
                    self._log.warning("ambiguity in recognition: %r" % output)
                    # FIXME: improve sorting criterion
                    results.sort(key=lambda result: 100 if result[0].has_dictation else 0)

                kaldi_rule, words = results[0]
                # FIXME: hack, but seems to work fine? only a problem for ambiguous rules containing dictation, which should be handled above
                words_are_dictation_mask = [True] * len(words)

        elif self._compiler.parsing_framework == 'token':
            kaldi_rule, words, words_are_dictation_mask = self._compiler.parse_output(output,
                dictation_info_func=lambda: (self.audio_store.current_audio_data, self._decoder.get_word_align(output)))
            if kaldi_rule is None:
                if words:
                    # We should never receive an unparsable recognition from kaldi, unless it's empty (from noise)
                    self._log.error("unable to parse recognition: %r" % output)
                return Recognition.construct_empty(self)

            if self._log.isEnabledFor(12):
                try:
                    self._log.log(12, "Alignment (word,time,length): %s", self._decoder.get_word_align(output))
                except KaldiError as e:
                    self._log.warning("Exception logging word alignment")

        else:
            raise EngineError("Invalid _compiler.parsing_framework")

        if not words:
            # Empty recognition, so bail before calling into dragonfly parsing/processing
            return Recognition.construct_empty(self)

        return Recognition(self, kaldi_rule=kaldi_rule, words=words, words_are_dictation_mask=words_are_dictation_mask)


#===========================================================================

class Recognition(object):
    """
    Kaldi recognition results class.
    """

    def __init__(self, engine, kaldi_rule, words, words_are_dictation_mask=None):
        assert isinstance(engine, KaldiEngine)
        self.engine = engine
        self.kaldi_rule = kaldi_rule
        self.words = tuple(words)
        if words_are_dictation_mask is None:
            assert not words
            words_are_dictation_mask = ()
        self.words_are_dictation_mask = tuple(words_are_dictation_mask)

        assert ((self.kaldi_rule and self.words and self.words_are_dictation_mask)
            or (not self.kaldi_rule and not self.words and not self.words_are_dictation_mask))
        self.parsed_output = ' '.join(words)
        self.has_dictation = any(words_are_dictation_mask)
        self.expected_error_rate = nan
        self.confidence = nan
        self.acceptable = None
        self.finalized = False

    @classmethod
    def construct_empty(cls, engine):
        return cls(engine, kaldi_rule=None, words=())

    # def __del__(self):
    #     # Exactly one of process() or fail() should be called by someone!
    #     if not self.finalized:
    #         self.engine._log.warning("%s not finalized!", self)
    #     # Note: this can be generated spurriously upon exit or testing

    def process(self, expected_error_rate=None, confidence=None, mimic=False):
        if expected_error_rate is not None: self.expected_error_rate = expected_error_rate
        if confidence is not None: self.confidence = confidence
        self.mimic = mimic
        self.acceptable = True
        assert self.words
        grammar_wrappers = self.engine._grammar_wrappers if mimic else self.engine._grammar_wrappers_copy
        grammar_wrapper = grammar_wrappers[id(self.kaldi_rule.parent_grammar)]
        with debug_timer(self.engine._log.debug, "dragonfly parse time"):
            grammar_wrapper.recognition_callback(self)
        self.finalized = True

    def fail(self, expected_error_rate=None, confidence=None, mimic=False):
        if expected_error_rate is not None: self.expected_error_rate = expected_error_rate
        if confidence is not None: self.confidence = confidence
        self.mimic = mimic
        self.acceptable = False
        self.engine._recognition_observer_manager.notify_failure(results=self)
        self.finalized = True


#===========================================================================

class GrammarWrapper(GrammarWrapperBase):

    def __init__(self, grammar, kaldi_rule_by_rule_dict, engine,
                 recobs_manager):
        GrammarWrapperBase.__init__(self, grammar, engine, recobs_manager)
        self.kaldi_rule_by_rule_dict = kaldi_rule_by_rule_dict

        self.active = True
        self.exclusive = False

    def phrase_start_callback(self, executable, title, handle):
        self.grammar.process_begin(executable, title, handle)

    def recognition_callback(self, recognition):
        words = recognition.words
        rule = recognition.kaldi_rule.parent_rule
        words_are_dictation_mask = recognition.words_are_dictation_mask
        try:
            assert (rule.active and rule.exported), "Kaldi engine should only ever return the correct rule"

            # Prepare the words and rule names for the element parsers
            rule_names = (rule.name,) + (('dgndictation',) if any(words_are_dictation_mask) else ())
            words_rules = tuple((word, 0 if not is_dictation else 1)
                for (word, is_dictation) in zip(words, words_are_dictation_mask))

            # Attempt to parse the recognition
            func = getattr(self.grammar, "process_recognition", None)
            if func:
                if not self._process_grammar_callback(func, words=words,
                                                      results=recognition):
                    # Return early if the method didn't return True or equiv.
                    return

            state = State(words_rules, rule_names, self.engine)
            state.initialize_decoding()
            for result in rule.decode(state):
                if state.finished():
                    root = state.build_parse_tree()
                    notify_args = (words, rule, root, recognition)
                    self.recobs_manager.notify_recognition(*notify_args)
                    with debug_timer(self.engine._log.debug, "rule execution time"):
                        rule.process_recognition(root)
                    self.recobs_manager.notify_post_recognition(*notify_args)
                    return

        except Exception as e:
            self.engine._log.error("Grammar %s: exception: %s" % (self.grammar._name, e), exc_info=True)

        # If this point is reached, then the recognition was not processed successfully
        self.engine._log.error("Grammar %s: failed to decode rule %s recognition %r." % (self.grammar._name, rule.name, words))

    # FIXME
    # def recognition_other_callback(self, StreamNumber, StreamPosition):
    #         func = getattr(self.grammar, "process_recognition_other", None)
    #         if func:
    #             func(words=False)
    #         return

    # FIXME
    # def recognition_failure_callback(self, StreamNumber, StreamPosition, Result):
    #         func = getattr(self.grammar, "process_recognition_failure", None)
    #         if func:
    #             func()
    #         return



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
This file offers access to various action classes.

"""

from .action_pause        import Pause
from .action_function     import Function
from .action_playback     import Playback
from .action_base         import (ActionBase, DynStrActionBase,
                                  Repeat, ActionError)
from .action_mimic        import Mimic
from .action_cmd          import RunCommand
from .action_context      import ContextAction
from .keyboard            import Keyboard, Typeable
from .action_key          import Key
from .action_text         import Text
from .action_paste        import Paste
from .action_mouse        import Mouse
from .action_waitwindow   import WaitWindow
from .action_focuswindow  import FocusWindow
from .action_startapp     import StartApp, BringApp
from .action_playsound    import PlaySound



#
# This file is part of Dragonfly.
# (c) Copyright 2020 by LexiconCode
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
    This module is a simple example of Dragonfly use.

    It shows how to use Dragonfly's Function action to call Python functions
    with the values associated with spoken Choice, Dictation and IntegerRef
    elements. This module can be activated in the same way as other
    Natlink macros by placing it in the "NatLink/MacroSystem" folder.

    This module requires the 'geocoder' Python package, which can be
    installed by running the following command:

      pip install geocoder

"""

import logging

import geocoder

from dragonfly import (Choice, Dictation, Function, Grammar, IntegerRef,
                       Key, MappingRule, Text, get_current_engine)


#---------------------------------------------------------------------------
# Define functions used by voice commands below.

def complete():
    print("Function is now complete.")


def cap_text(text):
    Text(text.upper()).execute()


def say_number(n):
    text = "I said number " + str(n)
    engine = get_current_engine()
    if engine:
        engine.speak(text)
    print(text)


def volume_control(n, volume_mode):
    for i in range(0, n):
        Key("volume" + volume_mode).execute()


data = {}
last_location = None
def get_location(location):
    global last_location
    last_location = location
    global data
    try:
        print(location)
        g = geocoder.osm(location.encode('utf-8'))
        data = g.json
    except (Exception) as e:
        log = logging.getLogger("location")
        log.exception("Error while getting data for location '%s': %s",
                      location, e)


def get_last_location_info(location_info):
    global last_location
    if not last_location:
        print('No location was specified. Say "get <location>" first.')
        return

    global data
    if not data:
        print("No data available for location '{0}'.".format(last_location))
    else:
        print(data.get(location_info))


def busy_func(fruit, animals, stuff, things):
    print(fruit)
    print(animals)
    print(stuff)
    print(things)


#---------------------------------------------------------------------------
# Create a mapping rule class which maps things you can say to actions.
#
# Note the relationship between the *mapping* and *extras* keyword
#  arguments.  The extras is a list of Dragonfly elements which are
#  available to be used in the specs of the mapping.  In this example
#  the IntegerRef("n") extra makes it possible to use "<n>" within a
#  mapping spec and pass an integer "n" to the associated action's
#  function.

class FunctionExamplesRule(MappingRule):
    mapping = {
        # Call the complete() function.
        "function complete": Function(complete),

        # Type capitalized text.
        "[dictation] capitalize <text>": Function(cap_text),

        # Say numbers.
        "say number <n>": Function(say_number),
        "say double number <n>": Function(lambda n: say_number(n * 2)),

        # Control the volume.
        "volume <volume_mode> [<n>]": Function(volume_control),

        # Retrieves address and country of <any location>.
        # 1. Say "get Yellowstone National Park".
        "get <location>":  Function(get_location),
        # The address and country of <any location> can be retrieved
        # afterwards.
        # 2. Say "what is the country".
        # 3. Say "what is the address".
        "what is [the] <location_info>": Function(get_last_location_info),

        # Print secret fruits and animals mapped to spelling words.
        # The values of 'stuff' and 'things' will also be printed.
        "secret <fruit> [<animals>]": Function(busy_func, stuff="shhhh...",
                                               things="more"),
    }

    extras = [
        IntegerRef("n", 1, 50),
        Dictation("text"),
        Dictation("location"),

        Choice("volume_mode", {
            "mute": "mute",
            "up": "up",
            "down": "down",
        }),

        Choice("fruit", {
            "alpha": "apple",
            "bravo": "blueberry",
            "charlie": "cherry",
        }),

        Choice("animals", {
            "delta": "dogs",
            "echo": "emus",
            "foxtrot": "foxes",
        }),

        Choice("location_info", {
            "address": "address",
            "country": "country",
        }),
    ]

    defaults = {
        "n": 1,
        "animals": "",
        "text": ""
    }


#---------------------------------------------------------------------------
# Create this module's grammar instance and add the rule to it.

grammar = Grammar("Function Examples")
grammar.add_rule(FunctionExamplesRule())


#---------------------------------------------------------------------------
# Load the grammar instance and define how to unload it.

grammar.load()

# Unload function which will be called by natlink at unload time.
def unload():
    global grammar
    if grammar:
        grammar.unload()
    grammar = None



#
# This file is part of Dragonfly.
# (c) Copyright 2018 by Dane Finlay
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

import locale
import logging
import sys
import time

from six import string_types, binary_type

import dragonfly.grammar.state as state_
from dragonfly import Window

from .recobs import TextRecobsManager
from ..base import (EngineBase, MimicFailure, ThreadedTimerManager,
                    DictationContainerBase, GrammarWrapperBase)


def _map_word(word):
    if isinstance(word, binary_type):
        word = word.decode(locale.getpreferredencoding())
    if word.isupper():
        # Convert dictation words to lowercase for consistent output.
        return word.lower(), 1000000
    else:
        return word, 0


class TextInputEngine(EngineBase):
    """Text-input Engine class. """

    _name = "text"
    DictationContainer = DictationContainerBase

    # -----------------------------------------------------------------------

    def __init__(self):
        EngineBase.__init__(self)
        self._language = "en"
        self._connected = False
        self._recognition_observer_manager = TextRecobsManager(self)
        self._timer_manager = ThreadedTimerManager(0.02, self)

    def connect(self):
        self._connected = True

    def disconnect(self):
        # Clear grammar wrappers on disconnect()
        self._grammar_wrappers.clear()
        self._connected = False

    # -----------------------------------------------------------------------
    # Methods for administrating timers.

    def create_timer(self, callback, interval, repeating=True):
        """
        Create and return a timer using the specified callback and repeat
        interval.

        Timers created using this engine will be run in a separate daemon
        thread, meaning that their callbacks will **not** be thread safe.
        :meth:`threading.Timer` may be used instead with no blocking issues.
        """
        return EngineBase.create_timer(self, callback, interval, repeating)

    # -----------------------------------------------------------------------
    # Methods for working with grammars.

    def _build_grammar_wrapper(self, grammar):
        return GrammarWrapper(grammar, self,
                              self._recognition_observer_manager)

    def _load_grammar(self, grammar):
        """ Load the given *grammar* and return a wrapper. """
        self._log.debug("Engine %s: loading grammar %s."
                        % (self, grammar.name))
        return self._build_grammar_wrapper(grammar)

    def _unload_grammar(self, grammar, wrapper):
        # No engine-specific unloading required.
        pass

    def activate_grammar(self, grammar):
        # No engine-specific grammar activation required.
        pass

    def deactivate_grammar(self, grammar):
        # No engine-specific grammar deactivation required.
        pass

    def activate_rule(self, rule, grammar):
        # No engine-specific rule activation required.
        pass

    def deactivate_rule(self, rule, grammar):
        # No engine-specific rule deactivation required.
        pass

    def update_list(self, lst, grammar):
        # No engine-specific list update is required.
        pass

    def set_exclusiveness(self, grammar, exclusive):
        wrapper = self._get_grammar_wrapper(grammar)
        if not wrapper:
            return

        wrapper.exclusive = exclusive

    # -----------------------------------------------------------------------
    # Miscellaneous methods.

    @classmethod
    def generate_words_rules(cls, words):
        # Convert words to Unicode, treat all uppercase words as dictation
        # words and other words as grammar words.
        # Minor note: this won't work for languages without capitalisation.
        return tuple(map(_map_word, words))

    def _do_recognition(self, delay=0):
        """
        Recognize words from standard input (stdin) in a loop until
        interrupted or :meth:`disconnect` is called.

        :param delay: time in seconds to delay before mimicking each
            command. This is useful for testing contexts.
        """
        self.connect()
        try:
            # Use iter to avoid a bug in Python 2.x:
            # https://bugs.python.org/issue3907
            for line in iter(sys.stdin.readline, ''):
                line = line.strip()
                if not line:  # skip empty lines.
                    continue

                # Delay between mimic() calls if necessary.
                if delay > 0:
                    time.sleep(delay)

                try:
                    self.mimic(line.split())
                except MimicFailure:
                    self._log.error("Mimic failure for words: %s", line)

                # Finish early if disconnect() was called.
                if self._connected is False:
                    break
        except KeyboardInterrupt:
            pass

    def mimic(self, words, **kwargs):
        """
        Mimic a recognition of the given *words*.

        :param words: words to mimic
        :type words: str|iter
        :Keyword Arguments:

           optional *executable*, *title* and/or *handle* keyword arguments
           may be used to simulate a specific foreground window context. The
           current foreground window attributes will be used instead for any
           keyword arguments not present.

        .. note::

           Any dictation words should be all uppercase, e.g. "HELLO WORLD".
           Dictation words not in uppercase will result in the engine
           **not** decoding and recognizing the command!
        """
        # Handle string input.
        if isinstance(words, string_types):
            words = words.split()

        # Don't allow non-iterable objects.
        if not iter(words):
            raise TypeError("%r is not a string or other iterable object"
                            % words)

        # Fail on empty input.
        if not words:
            raise MimicFailure("Invalid mimic input %r" % words)

        # Notify observers that a recognition has begun.
        self._recognition_observer_manager.notify_begin()

        # Generate the input for process_words.
        words_rules = self.generate_words_rules(words)

        w = Window.get_foreground()
        process_args = {
            "executable": w.executable,
            "title": w.title,
            "handle": w.handle,
        }
        # Allows optional passing of window attributes to mimic
        process_args.update(kwargs)

        # Call process_begin() for each grammar wrapper. Use a copy of
        # _grammar_wrappers in case it changes.
        for wrapper in self._grammar_wrappers.copy().values():
            wrapper.process_begin(**process_args)

        # Take another copy of _grammar_wrappers to use for processing.
        grammar_wrappers = self._grammar_wrappers.copy().values()

        # Count exclusive grammars.
        exclusive_count = 0
        for wrapper in grammar_wrappers:
            if wrapper.exclusive:
                exclusive_count += 1

        # Call process_words() for each grammar wrapper, stopping early if
        # processing occurred.
        processing_occurred = False
        for wrapper in grammar_wrappers:
            # Skip non-exclusive grammars if there are one or more exclusive
            # grammars.
            if exclusive_count > 0 and not wrapper.exclusive:
                continue

            # Process the grammar.
            processing_occurred = wrapper.process_words(words_rules)
            if processing_occurred:
                break

        # If no processing occurred, then the mimic failed.
        if not processing_occurred:
            self._recognition_observer_manager.notify_failure(None)
            raise MimicFailure("No matching rule found for words %r."
                               % (words,))

    def speak(self, text):
        self._log.warning("text-to-speech is not implemented for this "
                          "engine.")
        self._log.warning("Printing text instead.")
        print(text)

    @property
    def language(self):
        """
        Current user language of the SR engine.

        :rtype: str
        """
        return self._get_language()

    def _get_language(self):
        return self._language

    @language.setter
    def language(self, value):
        if not isinstance(value, string_types):
            raise TypeError("expected string, not %s" % value)

        self._language = value

    def _has_quoted_words_support(self):
        return False


class GrammarWrapper(GrammarWrapperBase):

    _log = logging.getLogger("engine")

    def __init__(self, grammar, engine, recobs_manager):
        GrammarWrapperBase.__init__(self, grammar, engine, recobs_manager)
        self.exclusive = False

    def process_begin(self, executable, title, handle):
        self.grammar.process_begin(executable, title, handle)

    def process_words(self, words):
        # Return early if the grammar is disabled or if there are no active
        # rules.
        if not (self.grammar.enabled and self.grammar.active_rules):
            return

        self._log.debug("Grammar %s: received recognition %r."
                        % (self.grammar.name, words))

        # TODO Make special grammar callbacks work properly.
        # These special methods are never called for this engine.
        results_obj = None
        if words == "other":
            func = getattr(self.grammar, "process_recognition_other", None)
            self._process_grammar_callback(func, words=words,
                                           results=results_obj)
            return
        elif words == "reject":
            func = getattr(self.grammar, "process_recognition_failure",
                           None)
            self._process_grammar_callback(func, results=results_obj)
            return

        # If the words argument was not "other" or "reject", then it is a
        # sequence of (word, rule_id) 2-tuples.
        words_rules = tuple(words)
        words = tuple(word for word, _ in words)

        # Call the grammar's general process_recognition method, if present.
        func = getattr(self.grammar, "process_recognition", None)
        if func:
            if not self._process_grammar_callback(func, words=words,
                                                  results=results_obj):
                # Return early if the method didn't return True or equiv.
                return

        # Iterate through this grammar's rules, attempting to decode each.
        # If successful, call that rule's method for processing the
        # recognition and return.
        s = state_.State(words_rules, self.grammar.rule_names, self.engine)
        for r in self.grammar.rules:
            if not (r.active and r.exported):
                continue
            s.initialize_decoding()
            for _ in r.decode(s):
                if s.finished():
                    try:
                        root = s.build_parse_tree()

                        # Notify observers using the manager *before*
                        # processing.
                        notify_args = (words, r, root, results_obj)
                        self.recobs_manager.notify_recognition(
                            *notify_args
                        )

                        r.process_recognition(root)

                        self.recobs_manager.notify_post_recognition(
                            *notify_args
                        )
                    except Exception as e:
                        self._log.exception("Failed to process rule "
                                            "'%s': %s" % (r.name, e))
                    return True

        self._log.debug("Grammar %s: failed to decode recognition %r."
                        % (self.grammar.name, words))
        return False



# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Example dragonfly RPC client.

This requires the 'requests' Python package.
"""

import json
import requests


def main():
    # Send a 'list_grammars' request to the server and print the response.
    # This is roughly what the 'dragonfly.rpc.util.send_rpc_request'
    # function does.
    url = "http://127.0.0.1:50051/jsonrpc"
    headers = {'content-type': 'application/json'}
    payload = {
        # Method to call.
        "method": "list_grammars",

        # Method parameters containing the server's security token.
        # Change the placeholder value to the security token normally
        # generated and printed when the RPC server is initialised.
        "params": {"security_token": "<security-token>"},
        # "params": ["<security-token>"],

        # Protocol version and request ID.
        "jsonrpc": "2.0",
        "id": 0,
    }
    response = requests.post(
        url, data=json.dumps(payload), headers=headers
    ).json()

    # Print the names of loaded grammars and rules.
    for grammar in response["result"]:
        rules = grammar["rules"]
        print("Grammar '%s' (%d rules)" % (grammar["name"], len(rules)))
        for rule in rules:
            print("Rule '%s'" % grammar["name"])


if __name__ == '__main__':
    main()



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Recognition observer class for the Natlink engine
============================================================================

"""

from ..base                import RecObsManagerBase
from ...grammar.grammar    import Grammar
from ...grammar.rule_base  import Rule
from ...grammar.elements   import Impossible


#---------------------------------------------------------------------------

class NatlinkRecObsManager(RecObsManagerBase):

    def __init__(self, engine):
        RecObsManagerBase.__init__(self, engine)
        self._grammar = None

        # Define a flag for limiting the number of notify_recognition()
        # calls to one per utterance / mimic call. This is necessary to
        # guarantee that observers are notified *before* rule processing and
        # are still notified about dictation and other grammar recognitions
        # via the special NatlinkRecObsGrammar grammar.
        self._complete_flag = False

    def notify_begin(self):
        super(NatlinkRecObsManager, self).notify_begin()
        self._complete_flag = False

    def notify_recognition(self, words, rule, root, results):
        if self._complete_flag:
            return

        super(NatlinkRecObsManager, self).notify_recognition(words, rule,
                                                             root, results)

    def notify_post_recognition(self, words, rule, root, results):
        if self._complete_flag:
            return

        super(NatlinkRecObsManager, self).notify_post_recognition(
            words, rule, root, results
        )
        self._complete_flag = True

    def _activate(self):
        if not self._grammar:
            self._grammar = NatlinkRecObsGrammar(self)
        self._grammar.load()

    def _deactivate(self):
        if self._grammar:
            self._grammar.unload()
        self._grammar = None


#---------------------------------------------------------------------------

class NatlinkRecObsGrammar(Grammar):

    def __init__(self, manager):
        self._manager = manager
        name = "_recobs_grammar"
        Grammar.__init__(self, name, description=None, context=None)

        rule = Rule(element=Impossible(), exported=True)
        self.add_rule(rule)

    #-----------------------------------------------------------------------
    # Callback methods for handling utterances and recognitions.

    def process_begin(self, executable, title, handle):
        self._manager.notify_begin()

    def process_recognition(self, words):
        raise RuntimeError("Recognition observer received an unexpected"
                           " recognition: %s" % (words,))

    def process_recognition_other(self, words, results):
        self._manager.notify_recognition(words, None, None, results)
        self._manager.notify_post_recognition(words, None, None, results)

    def process_recognition_failure(self, results):
        self._manager.notify_failure(results)



# coding=utf-8

import unittest
import string

from dragonfly.parsing.parse import spec_parser, CompoundTransformer
from dragonfly import Compound, Literal, Sequence, Optional, Empty, Alternative

# ===========================================================================

extras = {"an_extra": Alternative([Literal(u"1"), Literal(u"2")])}


def check_parse_tree(spec, expected):
    tree = spec_parser.parse(spec)
    output = CompoundTransformer(extras).transform(tree)
    assert output.element_tree_string() == expected.element_tree_string()
    return output


class TestLarkParser(unittest.TestCase):
    def test_literal(self):
        check_parse_tree("test   ", Literal(u"test"))

    def test_multiple_literals(self):
        check_parse_tree("test  hello world ", Literal(u"test hello world"))

    def test_parens(self):
        check_parse_tree("(test )   ", Literal(u"test"))

    def test_punctuation(self):
        check_parse_tree(",", Literal(u","))
        check_parse_tree("test's   ", Literal(u"test's"))
        check_parse_tree("cul-de-sac   ", Literal(u"cul-de-sac"))

    def test_sequence(self):
        check_parse_tree(
            " test <an_extra> [op]",
            Sequence([Literal(u"test"), extras["an_extra"], Optional(Literal(u"op"))]),
        )

    def test_alternative_no_parens(self):
        check_parse_tree(
            " test |[op] <an_extra>",
            Alternative(
                [
                    Literal(u"test"),
                    Sequence([Optional(Literal(u"op")), extras["an_extra"]]),
                ]
            ),
        )

    def test_alternative_parens(self):
        check_parse_tree(
            "( test |[op] <an_extra>)",
            Alternative(
                [
                    Literal(u"test"),
                    Sequence([Optional(Literal(u"op")), extras["an_extra"]]),
                ]
            ),
        )

    def test_optional_alternative(self):
        check_parse_tree("[test|test's]", Optional(Alternative([Literal(u"test"), Literal(u"test's")])))

    def test_digit_in_word(self):
        check_parse_tree("F2", Literal(u"F2"))

    def test_unicode(self):
        check_parse_tree(u"touché", Literal(u"touché"))

    def test_bool_special_in_sequence(self):
        output = check_parse_tree(
            " test <an_extra> [op] {test_special}",
            Sequence([Literal(u"test"), extras["an_extra"], Optional(Literal(u"op"))]),
        )
        assert output.test_special == True
        assert all(getattr(child, 'test_special', None) == None for child in output.children)

    def test_other_special_in_sequence(self):
        output = check_parse_tree(
            " test <an_extra> [op] {test_special=4}",
            Sequence([Literal(u"test"), extras["an_extra"], Optional(Literal(u"op"))]),
        )
        assert output.test_special == 4
        assert all(getattr(child, 'test_special', None) == None for child in output.children)

    def test_bool_special_in_alternative(self):
        output = check_parse_tree(
            "foo | bar {test_special} | baz",
            Alternative([
                Literal(u"foo"),
                Literal(u"bar"),
                Literal(u"baz"),
            ]),
        )
        assert getattr(output.children[0], 'test_special', None) == None
        assert output.children[1].test_special == True
        assert getattr(output.children[2], 'test_special', None) == None

    def test_other_special_in_alternative(self):
        output = check_parse_tree(
            "foo | bar {test_special=4} | baz",
            Alternative([
                Literal(u"foo"),
                Literal(u"bar"),
                Literal(u"baz"),
            ]),
        )
        assert getattr(output.children[0], 'test_special', None) == None
        assert output.children[1].test_special == 4
        assert getattr(output.children[2], 'test_special', None) == None


# ===========================================================================

if __name__ == "__main__":
    unittest.main()
